from collections import OrderedDict
from flask import Blueprint, g, _request_ctx_stack, current_app, request
from werkzeug import LocalProxy
from operator import attrgetter
from types import FunctionType
from functools import partial

fs = LocalProxy(lambda: current_app.extensions['flarf'].filters)

class FlarfFilter(object):
    """
    A class used to instance a filter result on application request

    :param filter_tag:          The name of the filter
    :param filter_precedence:   If you need to order your filters in some way,
                                set this as an integer, defaults to 100.
                                Filters will be ordered from smallest number to
                                largest
    :param filter_params:       A list of paramaters used by the filter on request.
                                May be strings or functions. Strings must be
                                attributes of request, found within (request.values,
                                request.form, or request.files) or reference a
                                method on the filter class that take request as
                                an argument. Passed in functions must take request
                                as a single argument.
                                On filter run params are run in order of:
                                   - if a function
                                   - if an attribute of flask request
                                   - if in request.values, request.form,
                                     and request.files
                                If not found, the param will be set on the filter
                                as an attr of None.
    :param filter_on:           A list of routes to use the filter on, default
                                is ['all'], except static routes
    :param filter_skip:         A list routes to pass and not use filter.
                                By default, all static routes are skipped
    """
    def __init__(self,
                 filter_tag,
                 filter_precedence=100,
                 filter_params=None,
                 filter_on=['all'],
                 filter_skip=None):
        self.filter_tag = filter_tag
        self.filter_precedence = filter_precedence
        self.filter_params = self.set_params(filter_params)
        self.filter_on = filter_on
        self.filter_pass = ['static']
        if filter_skip:
            self.filter_pass.extend(filter_skip)

    def set_params(self, params):
        d = OrderedDict()
        for p in params:
            if isinstance(p, FunctionType):
                d[p.__name__] = p
            elif p[:4] == 'get_':
                d[p[4:]] = partial(getattr(self, p))
            elif hasattr(request, p):
                d[p] = partial(getattr(self, 'param_request'), p)
            else:
                d[p] = partial(getattr(self, 'param_param'), p)
        return d

    def get_ctx_prc(self):
        def ctx_prc(tag):
            return getattr(g, tag, None)
        return {self.filter_tag: ctx_prc(self.filter_tag)}

    def param_request(self, param, request):
        return getattr(request, param)

    def param_param(self, param, request):
        p = filter(None, [request.values.get(param, None),
                          request.view_args.get(param, None),
                          request.files.get(param, None)])
        if p:
            return p.pop()
        else:
            return None

    def filter_by_param(self, request):
        for k, v in self.filter_params.items():
            setattr(self, k, v(request))

    def filter_request(self, request):
        self.filter_by_param(request)
        setattr(g, self.filter_tag, self)


class Flarf(object):
    """
    The Flarf extension object to registered with a Flask application.

    :param app:                 The application to register the function on.
    :param before_request_func: The before request function to run the filters.
                                Defaults to self.flarf_run_filters
    :param filter_cls:          The class used as a filter, defaults to
                                FlarfFilter, used when receiving dicts as filters
    :param filters:             A list of filter instances(or dicts mappable
                                to filter_cls instances) to be run per request.
    """
    def __init__(self,
                 app=None,
                 before_request_func=None,
                 filter_cls=FlarfFilter,
                 filters=None):
        self.app = app
        self.before_request_func = self.set_before_request_func(before_request_func)
        self.filter_cls = filter_cls
        self.filters = self.process_filters(filters)

        if app is not None:
            self.app = app
            self.init_app(self.app)
        else:
            self.app = None

    def set_before_request_func(self, before_request_func):
        if before_request_func:
            return before_request_func
        else:
            return self.flarf_run_filters

    def process_filters(self, filters):
        d = OrderedDict()
        fs = self.check_filters(filters)
        ofs = self.order_filters(fs)
        for f in ofs:
            d[f.filter_tag] = f
        return d

    def check_filters(self, filters):
        return [self.reflect_filter(f) for f in filters]

    def reflect_filter(self, afilter):
        if isinstance(afilter, dict):
            return self.filter_cls(**afilter)
        else:
            return afilter

    def order_filters(self, filters):
        return sorted(filters, key=attrgetter('filter_precedence'))

    def init_context_processors(self, app):
        for f in self.filters.values():
           app.context_processor(f.get_ctx_prc)

    def init_app(self, app):
        app.before_request(self.before_request_func)
        self.init_context_processors(app)
        app.extensions['flarf'] = self

    def flarf_run_filters(self):
        """
        A before_request function registered on the extension that runs each filter.
        """
        r = _request_ctx_stack.top.request
        if r.url_rule:
            request_endpoint = str(r.endpoint).rsplit('.')[-1]
            for f in self.filters.values():
                if request_endpoint not in f.filter_pass:
                    if request_endpoint or 'all' in f.filter_on:
                        rv = f.filter_request(r)
                        if rv:
                            return rv



"""
Flarf: Flask Request Filter
-------------
Configurable request filters
"""

from setuptools import setup

setup(
    name='Flask-Flarf',
    version='0.0.4',
    url='https://github.com/thrisp/flarf',
    license='MIT',
    author='Thrisp/Hurrata',
    author_email='blueblank@gmail.com',
    description='Flask request filtering',
    long_description=__doc__,
    packages=['flask_flarf'],
    zip_safe=False,
    platforms='any',
    install_requires=[
        'Flask>=0.9'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from __future__ import with_statement
import sys
import os
from flask import Flask, render_template, current_app, g, request, redirect
from flask.ext.flarf import Flarf, FlarfFilter, flarf
import unittest

class FlarfTestCase(unittest.TestCase):
    def setUp(self):
        def custom_before_func():
            setattr(g, 'custom_before_func', True)
        class CustomFilter(FlarfFilter):
            def __init__(self, something_custom="NOTHING", **kwargs):
                self.something_custom = something_custom
                super(CustomFilter, self).__init__(**kwargs)
            def filter_request(self, request):
                setattr(g, 'custom_filter_run', True)
                setattr(g, self.filter_tag, self.something_custom)
                return redirect('/')
        def path_to_upper(request):
            return request.path.upper()
        self.custom_before_func = custom_before_func
        self.custom_filter = CustomFilter
        test_filter1 = FlarfFilter(filter_tag='test_filter1',
                                   filter_precedence=100,
                                   filter_params=['path', path_to_upper],
                                   filter_on=['includeme'])
        test_filter2 = FlarfFilter(filter_tag='test_filter2',
                                   filter_precedence=200,
                                   filter_params=['values', 'yod', 'zed'])
        test_filter3 = {'filter_tag': 'test_filter3',
                        'filter_precedence': 300,
                        'filter_params': ['path', 'args'],
                        'filter_skip':['passme']}
        test_filter4 = CustomFilter(filter_tag='test_filter4',
                                    filter_precedence=100,
                                    filter_params=['path', path_to_upper],
                                    filter_on=['app_route'],
                                    something_custom="I am test filter 4")
        test_filter5 = {'filter_tag': 'test_filter5',
                        'filter_precedence': 300,
                        'filter_params': ['path', 'args'],
                        'filter_skip':['passme'],
                        'something_custom': 'I am test filter 5'}
        self.test_filters1 = [test_filter1]
        self.test_filters2 = [test_filter1, test_filter2, test_filter3]
        self.test_filters3 = [test_filter4, test_filter5]
        pre_app = Flask(__name__)
        @pre_app.route('/')
        def test_index():
            return g.__dict__
        @pre_app.route('/app_route')
        def test_app_route():
            return g.__dict__
        @pre_app.route('/context_processor')
        def test_c_route():
            return render_template('test_template.html')
        post_app = Flask(__name__)
        @post_app.route('/')
        def test_index():
            return g.__dict__
        Flarf(post_app, filters=self.test_filters1)
        self.base_app = post_app
        self.pre_app = pre_app

    def tearDown(self):
        self.pre_app = None

    def test_base(self):
        with self.base_app.test_request_context('/'):
            self.base_app.preprocess_request()
            self.assertIsNotNone(self.base_app.extensions['flarf'])
            self.assertIsNotNone(getattr(g, 'test_filter1', None))

    def test_filters(self):
        Flarf(self.pre_app, filters=self.test_filters2)
        with self.pre_app.test_request_context('/app_route?zed=z'):
            self.pre_app.preprocess_request()
            self.assertIsNotNone(getattr(g, 'test_filter3', None))
            self.assertEqual(getattr(g, 'test_filter3', None),
                             getattr(g, 'test_filter3', None))
            self.assertEqual(g.test_filter1.path_to_upper, u'/APP_ROUTE')
            #self.assertEqual(u'/app_route', g.test_filter3.path)
            self.assertEqual(g.test_filter1.path,
                             g.test_filter3.path)

    def test_context_processor(self):
        Flarf(self.pre_app, filters=self.test_filters2)
        with self.pre_app.test_client() as ct:
            rv = ct.get('/context_processor?zed=z')
            self.assertIsNotNone(rv.data)
            self.assertEqual(rv.data.decode(), g.test_filter2.zed)

    def test_custom_before_request_func(self):
        Flarf(self.pre_app,
              filters=self.test_filters2,
              before_request_func=self.custom_before_func)
        with self.pre_app.test_request_context('/'):
            self.pre_app.preprocess_request()
            self.assertTrue(g.custom_before_func)

    def test_custom_filter_cls(self):
        Flarf(self.pre_app, filter_cls=self.custom_filter, filters=self.test_filters3)
        with self.pre_app.test_request_context('/'):
            self.pre_app.preprocess_request()
            self.assertTrue(g.custom_filter_run)
        with self.pre_app.test_client() as ct:
            rv = ct.get('/app_route')
            self.assertEqual(rv.status_code, 302)

    def test_include_route(self):
        @self.pre_app.route('/includeme')
        def includeme():
            return g.__dict__
        Flarf(self.pre_app, filters=self.test_filters2)
        with self.pre_app.test_request_context('/includeme'):
            self.pre_app.preprocess_request()
            #self.assertTrue(g.test_filter1.path)
            #self.assertEqual(g.test_filter1.path, u'/includeme')

    def test_exclude_route(self):
        @self.pre_app.route('/passme')
        def passme():
            return g.__dict__
        Flarf(self.pre_app, filters=self.test_filters2)
        with self.pre_app.test_request_context('/passme'):
            self.pre_app.preprocess_request()
            with self.assertRaises(AttributeError):
                g.test_filter3

if __name__ == '__main__':
    unittest.main()
